from decimal import Decimal
from typing import Literal, Tuple


def decode_uri(currency: Literal["btc", "bch", "bsv"], uri: str) -> Tuple[str, int]:
    """
    decode_uri: Converts URI to address and Satoshis, validates against
    currency.

    amount is called amount because we may add a currency that doesn't have
    'satoshis' one day. This is meant to be normalizing.

    This does *not* validate address.
    """
    if "&" in uri:
        raise ValueError("& in URI not supported.")

    uri_currency = uri.split(":")[0]
    uri_address = uri.split(":")[1].split("?")[0]
    uri_amount = Decimal(uri.split("=")[1])

    address = uri_address
    satoshis = int(uri_amount * 100000000)

    if currency == "btc":
        if uri_currency != "bitcoin":
            raise ValueError("URI does not start with bitcoin: for BTC")
    elif currency == "bch":
        if uri_currency != "bitcoincash":
            raise ValueError("URI does not start with bitcoincash: for BCH")
        else:
            address = uri.split("?")[0]
    elif currency == "bsv":
        if uri_currency != "bitcoin":
            raise ValueError("URI does not start with bitcoin: for BSV")
    else:
        raise ValueError("Currency must be one of: btc, bch, bsv")

    return (address, satoshis)



"""
WalkingLiberty

Sort of a Bitcoin, Bitcoin Cash, and Bitcoin SV wallet API
"""

from hashlib import sha256
from importlib.metadata import version
from typing import Literal, Optional, Union

import bit
import bitcash
import bitsv

from walkingliberty.utilities import decode_uri

__version__ = version(__package__)

VALID_CURRENCIES = ("btc", "bch", "bsv")
VALID_WALLET_MODES = ("deterministic-type1", "wif", "address")


class WalkingLiberty:
    """
    WalkingLiberty class.

    private_key is address when wallet_mode is set to address.
    """

    def __init__(
        self,
        currency: Literal["btc", "bch", "bsv"],
        wallet_mode: Literal["deterministic-type1", "wif", "address"],
        private_key: str,
    ) -> None:
        if currency not in VALID_CURRENCIES:
            message = "currency must be one of: {}".format(VALID_CURRENCIES)
            raise ValueError(message)

        if wallet_mode not in VALID_WALLET_MODES:
            message = "wallet_mode must be one of: {}".format(VALID_WALLET_MODES)
            raise ValueError(message)

        if currency == "btc":
            self.bit = bit
        elif currency == "bch":
            self.bit = bitcash
        elif currency == "bsv":
            self.bit = bitsv

        self.currency = currency
        self.wallet_mode: str = wallet_mode
        self._satoshis: Optional[int] = None

        if self.wallet_mode == "deterministic-type1":
            private_key_bytes = bytes(private_key, "utf-8")
            self.private_key = self.bit.Key.from_hex(
                sha256(private_key_bytes).hexdigest()
            )
        elif self.wallet_mode == "wif":
            self.private_key = self.bit.Key(private_key)

        if self.wallet_mode == "address":
            self._address = private_key
        else:
            address = self.private_key.address
            assert isinstance(address, str), "address must be string"
            self._address = address

    def address(self) -> str:
        """
        Returns address for key
        """
        return self._address

    def balance(
        self, unit: str = "satoshi", use_cache: bool = False
    ) -> Union[int, str]:
        """
        Returns balance for a private key's address.

        Optionally, set unit for desired unit. (satoshi, btc, bsv, usd, etc)

        use_cache: Controls whether we cache the balance or not.

        Let's say you wanted to get balance in satoshis and USD,
        this would be only query the Blockchain APIs once. But, if
        you want to monitor balance over time on the same WalkingLiberty
        object, the cache would get the starting balance, only.
        """
        if self.currency == "bsv":
            NetworkAPI = self.bit.network.NetworkAPI("main")
        else:
            NetworkAPI = self.bit.network.NetworkAPI

        if use_cache:
            if self._satoshis is None:
                self._satoshis = NetworkAPI.get_balance(self._address)

            satoshis = self._satoshis
        else:
            satoshis = NetworkAPI.get_balance(self._address)

        satoshi_to_currency = self.bit.network.satoshi_to_currency_cached
        balance = satoshi_to_currency(satoshis, unit)

        # Return int if satoshis, str if anything else.
        if unit == "satoshi":
            return int(balance)
        else:
            assert isinstance(balance, str), "balance must be str."
            return balance

    def send(self, address: str, satoshis: int, unit: str = "satoshi") -> str:
        """
        Sends amount to address from the private_key's wallet.

        Returns TXID.
        """
        if self.wallet_mode == "address":
            raise ValueError("Cannot send without private key.")

        outputs = [(address, satoshis, unit)]
        # Have to call get_unspents() or get_balance() before send()
        self.private_key.get_unspents()
        txid = self.private_key.send(outputs, combine=False)
        assert isinstance(txid, str), "TXID must be str."
        return txid

    def pay(self, uri: str) -> str:
        """
        Pays URI.

        Returns TXID.
        """
        if self.wallet_mode == "address":
            raise ValueError("Cannot pay without private key.")

        address, satoshis = decode_uri(self.currency, uri)
        outputs = [(address, satoshis, "satoshi")]
        # Have to call get_unspents() or get_balance() before send()
        self.private_key.get_unspents()
        txid = self.private_key.send(outputs, combine=False)
        assert isinstance(txid, str), "TXID must be str."
        return txid

    def sweep(self, address: str) -> str:
        """
        Sweeps all coins from private_key into address.

        Returns TXID.
        """
        if self.wallet_mode == "address":
            raise ValueError("Cannot sweep without private key.")

        self.private_key.get_unspents()
        txid = self.private_key.send([], combine=True, leftover=address)
        assert isinstance(txid, str), "TXID must be str."
        return txid



"""
WalkingLiberty CLI
"""

import argparse

import segno

import walkingliberty


def address(args: argparse.Namespace) -> None:
    WalkingLiberty = walkingliberty.WalkingLiberty(
        args.currency, args.wallet_mode, args.phrase
    )
    address = WalkingLiberty.address()
    if args.qr is True:
        segno.make(address).terminal()
    print(address)


def private_key(args: argparse.Namespace) -> None:
    """Display the private key"""
    WalkingLiberty = walkingliberty.WalkingLiberty(
        args.currency, args.wallet_mode, args.phrase
    )
    print(WalkingLiberty.private_key.to_wif())


def balance(args: argparse.Namespace) -> None:
    WalkingLiberty = walkingliberty.WalkingLiberty(
        args.currency, args.wallet_mode, args.phrase
    )
    print(WalkingLiberty.balance(args.unit))


def send(args: argparse.Namespace) -> None:
    WalkingLiberty = walkingliberty.WalkingLiberty(
        args.currency, args.wallet_mode, args.phrase
    )
    print(WalkingLiberty.send(args.address, args.satoshis, args.unit))


def pay(args: argparse.Namespace) -> None:
    WalkingLiberty = walkingliberty.WalkingLiberty(
        args.currency, args.wallet_mode, args.phrase
    )
    print(WalkingLiberty.pay(args.uri))


def sweep(args: argparse.Namespace) -> None:
    WalkingLiberty = walkingliberty.WalkingLiberty(
        args.currency, args.wallet_mode, args.phrase
    )
    print(WalkingLiberty.sweep(args.address))


def main() -> None:
    parser = argparse.ArgumentParser(description="WalkingLiberty CLI.")
    parser.add_argument(
        "--currency", help="Currency: btc, bch, or bsv", type=str.lower, default="btc"
    )
    parser.add_argument(
        "--wallet-mode",
        help="Wallet mode: deterministic-type1, wif, or address",
        default="deterministic-type1",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"WalkingLiberty {walkingliberty.__version__}",
    )

    subparser = parser.add_subparsers()
    address_subparser = subparser.add_parser(
        "address", help="Returns phrase's address."
    )
    address_subparser.set_defaults(func=address)
    address_subparser.add_argument("phrase", help="Deterministic phrase.")
    address_subparser.add_argument(
        "--qr", help="QR code", action="store_true", default=False
    )

    private_key_subparser = subparser.add_parser(
        "private-key", help="Returns private key."
    )
    private_key_subparser.set_defaults(func=private_key)
    private_key_subparser.add_argument("phrase", help="Deterministic phrase.")

    balance_subparser = subparser.add_parser(
        "balance", help="Returns phrase's balance."
    )
    balance_subparser.add_argument("--unit", help="As unit/currency", default="satoshi")
    balance_subparser.set_defaults(func=balance)
    balance_subparser.add_argument("phrase", help="Deterministic phrase.")

    formatter_class = argparse.ArgumentDefaultsHelpFormatter
    send_subparser = subparser.add_parser(
        "send", help="Sends cryptocurrency.", formatter_class=formatter_class
    )
    send_subparser.set_defaults(func=send)
    help = "Deterministic phrase."
    send_subparser.add_argument("phrase", help=help)
    help = "Address to send to."
    send_subparser.add_argument("address", help=help)
    help = "Satoshis to send"
    send_subparser.add_argument("satoshis", help=help, type=int)
    help = "As unit/currency"
    send_subparser.add_argument("--unit", help=help, default="satoshi")

    pay_subparser = subparser.add_parser(
        "pay", help="Pay a URI.", formatter_class=formatter_class
    )
    pay_subparser.set_defaults(func=pay)
    pay_subparser.add_argument("phrase", help="Deterministic phrase.")
    pay_subparser.add_argument("uri", help="URI to pay.")

    sweep_subparser = subparser.add_parser("sweep", help="Sweeps all cryptocurrency.")
    help = "Deterministic phrase."
    sweep_subparser.add_argument("phrase", help=help)
    help = "Address to send to."
    sweep_subparser.add_argument("address", help=help)
    sweep_subparser.set_defaults(func=sweep)

    args = parser.parse_args()
    # This calls the function or wrapper function, depending on what we set
    # above.
    if "func" in args:
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()
